#!BPY

"""
Name: '3D Studio'
Blender: 232
Group: 'Export'
Tip: 'Export to 3DS file format. (.3ds)'
"""



######################################################
# 3ds Importer
# By:  Bob Holcomb
# Date: 26 Jan 04
# Ver: 0.5
######################################################
# This script imports a 3ds file and the materials
#  into blender for editing.  Hopefully
# this will make it into a future version as an import
# feature of the menu.  Loader is based on 3ds loader
# from www.gametutorials.com(Thanks DigiBen).
######################################################

######################################################
# Importing modules
######################################################

import Blender
from Blender import NMesh, Scene, Object, Material
from Blender.BGL import *
from Blender.Draw import *
from Blender.Window import *
from Blender.Image import *
from Blender.Material import *

import sys, struct, string, types
from types import *

import os
from os import path


######################################################
# Data Structures
######################################################

#Some of the chunks that we will see
#----- Primary Chunk, at the beginning of each file
PRIMARY=19789 				#0x4D4D

#------ Main Chunks
OBJECTINFO=15677				#0x3D3D			// This gives the version of the mesh and is found right before the material and object information
VERSION=2					#0x0002			// This gives the version of the .3ds file
EDITKEYFRAME=45056			#0xB000			// This is the header for all of the key frame info

#------ sub defines of OBJECTINFO
MATERIAL=45055				#0xAFFF			// This stored the texture info
OBJECT=16384					#0x4000			// This stores the faces, vertices, etc...

#------ sub defines of MATERIAL
MATNAME=40960				#0xA000			// This holds the material name
MATDIFFUSE=40992				#0xA020			// This holds the color of the object/material
MATMAP=41472				#0xA200			// This is a header for a new material
MATMAPFILE=41728				#0xA300			// This holds the file name of the texture

OBJECT_MESH=16640			#0x4100			// This lets us know that we are reading a new object

#------ sub defines of OBJECT_MESH
OBJECT_VERTICES=16656			#0x4110			// The objects vertices
OBJECT_FACES=16672			#0x4120			// The objects faces
OBJECT_MATERIAL=16688		#0x4130			// This is found if the object has a material, either texture map or color
OBJECT_UV=16704				#0x4140			// The UV texture coordinates
OBJECT_TRANS_MATRIX=16736	#0x4160			// The translation matrix of the object 54 bytes

#may want to add light, camera, keyframe chunks.
class vert_chunk:
	verts=[]
	size=0
	ID=OBJECT_VERTICES	#0x4110
	num_verts=0


	def __init__(self):
		self.size=0
		self.verts=[]
		self.ID=OBJECT_VERTICES

	def get_size(self):
		temp_size=0
		temp_size=6 #bytes of this header
		temp_size+=2 #for the number of verts short
		for vert in self.verts:
			temp_size+=12  #3 floats x 4 bytes each
		self.size=temp_size
		print "vert_chunk size: ", self.size
		return self.size

	def write(self, file):
		binary_format="<HIH"  #all header info looks like this
		#calculate size
		temp_data=[0]*3
		temp_data[0]=self.ID
		temp_data[1]=self.size
		temp_data[2]=len(self.verts)
		#write header
		data=struct.pack(binary_format,temp_data[0],temp_data[1], temp_data[2])
		file.write(data)
		#write verts
		for vert in self.verts:
			binary_format="<3f"
			temp_data=[0.0]*3
			temp_data[0]=vert[0]
			temp_data[1]=vert[1]
			temp_data[2]=vert[2]
			data=struct.pack(binary_format,temp_data[0],temp_data[1], temp_data[2])
			file.write(data)

class face_chunk:
	faces=[]
	size=0
	ID=OBJECT_FACES   #0x4120
	num_faces=0

	def __init__(self):
		self.size=0
		self.faces=[]
		self.ID=OBJECT_FACES
		self.num_faces=0

	def get_size(self):
		temp_size=0
		temp_size=6 #bytes of this header
		temp_size+=2 #num faces info
		for face in self.faces:
			temp_size+=8  #4 short ints x 2 bytes each
		self.size=temp_size
		print "face_chunk size: ", self.size
		return self.size

	def write(self, file):
		binary_format="<HIH"  #all header info looks like this
		#calculate size
		temp_data=[0]*3
		temp_data[0]=self.ID
		temp_data[1]=self.size
		temp_data[2]=len(self.faces)

		#write header
		data=struct.pack(binary_format,temp_data[0],temp_data[1], temp_data[2])
		file.write(data)
		#write faces
		for face in self.faces:
			binary_format="<4H"
			temp_data=[0]*4
			temp_data[0]=face[0]
			temp_data[1]=face[1]
			temp_data[2]=face[2]
			temp_data[3]=0 #only used by 3d studio
			data=struct.pack(binary_format,temp_data[0],temp_data[1], temp_data[2], temp_data[3])
			file.write(data)

#this one is a bit broken, need to get a face list using materials and write that too
class obj_material_chunk:
	name=""
	size=0
	ID=OBJECT_MATERIAL  #0x4130

	def __init__(self):
		self.size=0
		self.name=""
		self.ID=OBJECT_MATERIAL

	def get_size(self):
		temp_size=0
		temp_size=6 #bytes of this header
		temp_size+=(len(self.name)+1)
		self.size=temp_size
		print "obj material chunk size: ", self.size
		return self.size

	def write(self, file):
		binary_format="<HI"  #all header info looks like this
		#calculate size
		temp_data=[0]*2
		temp_data[0]=self.ID
		temp_data[1]=self.size
		#write header
		data=struct.pack(binary_format,temp_data[0],temp_data[1])
		file.write(data)
		#write name
		name_length=len(self.name)+1
		binary_format="<"+str(name_length)+"s"
		data=struct.pack(binary_format, self.name)
		file.write(data)

class uv_chunk:
	uv=[]
	size=0
	ID=OBJECT_UV
	num_uv=0

	def __init__(self):
		self.size=0
		self.uv=[]
		self.ID=OBJECT_UV
		self.num_uv=0

	def get_size(self):
		temp_size=0
		temp_size=6 #bytes of this header
		temp_size+=2 #for num UV
		for this_uv in self.uv:
			temp_size+=8  #2 floats at 4 bytes each
		self.size=temp_size
		print "uv chunk size: ", self.size
		return self.size

	def write(self, file):
		binary_format="<HIH"  #all header info looks like this
		#calculate size
		temp_data=[0]*3
		temp_data[0]=self.ID
		temp_data[1]=self.size
		temp_data[2]=len(self.uv)
		#write header
		data=struct.pack(binary_format,temp_data[0],temp_data[1], temp_data[2])
		file.write(data)
		#write verts
		for this_uv in self.uv:
			binary_format="<2f"
			temp_data=[0.0]*2
			temp_data[0]=this_uv[0]
			temp_data[1]=this_uv[1]
			data=struct.pack(binary_format,temp_data[0],temp_data[1])
			file.write(data)

class mesh_chunk:
	v_chunk=vert_chunk()
	f_chunk=face_chunk()
	m_chunk=obj_material_chunk()
	uv_chunk=uv_chunk()
	size=0
	ID=OBJECT_MESH

	def __init__(self):

		self.v_chunk=vert_chunk()
		self.f_chunk=face_chunk()
		self.m_chunk=obj_material_chunk()
		self.uv_chunk=uv_chunk()
		self.size=0
		self.ID=OBJECT_MESH

	def get_size(self):
		temp_size=0
		temp_size=6 #bytes of this header
		temp_size+=self.v_chunk.get_size()
		temp_size+=self.f_chunk.get_size()
		#temp_size+=self.m_chunk.get_size()
		temp_size+=self.uv_chunk.get_size()
		self.size=temp_size
		print "object mesh chunk size: ", self.size
		return self.size

	def write(self, file):
		binary_format="<HI"  #all header info looks like this
		#calculate size
		temp_data=[0]*2
		temp_data[0]=self.ID
		temp_data[1]=self.size
		#write header
		data=struct.pack(binary_format,temp_data[0],temp_data[1])
		file.write(data)
		#write stuff
		self.v_chunk.write(file)
		self.f_chunk.write(file)
		#self.m_chunk.write(file)
		self.uv_chunk.write(file)

class object_chunk:
	name=""
	mesh_chunks=[]
	size=0
	ID=OBJECT

	def __init__(self):
		self.name=""
		self.mesh_list=[]
		self.size=0
		self.ID=OBJECT

	def get_size(self):
		temp_size=0
		temp_size=6 #bytes of this header
		temp_size+=len(self.name)+1 #+1 for null character
		for mesh in self.mesh_chunks:
			temp_size+=mesh.get_size()
		self.size=temp_size
		print "object chunk size: ", self.size
		return self.size

	def write(self, file):
		binary_format="<HI"  #all header info looks like this
		#calculate size
		temp_data=[0]*2
		temp_data[0]=self.ID
		temp_data[1]=self.size
		#write header and version
		data=struct.pack(binary_format,temp_data[0],temp_data[1])
		file.write(data)
		#write name
		name_length=len(self.name)+1
		binary_format="<"+str(name_length)+"s"
		data=struct.pack(binary_format, self.name)
		file.write(data)
		#write stuff
		for mesh in self.mesh_chunks:
			mesh.write(file)

class object_info_chunk:
	obj_chunks=[]
	#mat_chunks=[]
	size=0
	ID=OBJECTINFO

	def __init__(self):
		self.obj_chunks=[]
		self.ID=OBJECTINFO
		self.size=0

	def get_size(self):
		temp_size=0
		temp_size=6 #header info
		for obj in self.obj_chunks:
			temp_size+=obj.get_size()
		self.size=temp_size
		print "object info size: ", self.size
		return self.size

	def write(self, file):
		binary_format="<HI"  #all header info looks like this
		#calculate size
		temp_data=[0]*2
		temp_data[0]=self.ID
		temp_data[1]=self.size
		#write header and version
		data=struct.pack(binary_format,temp_data[0],temp_data[1])
		file.write(data)
		#write all the objects
		for obj in self.obj_chunks:
			obj.write(file)



class version_chunk:
	ID=VERSION  #0x002
	size=0
	version=3

	def __init__(self):
		self.ID=VERSION
		self.size=0
		self.version=3 #that the document that I'm using

	def get_size(self):
		temp_size=0
		temp_size+=6  #bytes of this header
		temp_size+=4 #bytes for the version info
		self.size=temp_size
		print "version chunk size: ", self.size
		return self.size

	def write(self, file):
		binary_format="<HII"  #all header info looks like this
		#calculate size
		temp_data=[0]*3
		temp_data[0]=self.ID
		temp_data[1]=self.size
		temp_data[2]=self.version
		#write header and version
		data=struct.pack(binary_format,temp_data[0],temp_data[1], temp_data[2])
		file.write(data)

class primary_chunk:
	version=None
	obj_info=None
	size=0
	ID=PRIMARY #0x4D4D

	def __init__(self):
		self.version=version_chunk()
		self.obj_info=object_info_chunk()
		self.size=0
		self.ID=PRIMARY

	def get_size(self):
		temp_size=0
		temp_size=6 #bytes of this header
		temp_size+=self.version.get_size()
		temp_size+=self.obj_info.get_size()
		self.size=temp_size
		print "primary chunk size: ", self.size
		return self.size

	def write(self, file):
		binary_format="<HI"  #all header info looks like this
		#calculate size
		temp_data=[0]*2
		temp_data[0]=self.ID
		temp_data[1]=self.size
		#write header
		data=struct.pack(binary_format,temp_data[0],temp_data[1])
		file.write(data)
		#write version chunk
		self.version.write(file)
		#write object_info chunk
		self.obj_info.write(file)

#the chunk class
class chunk:
	ID=0
	length=0
	bytes_read=0

	#we don't read in the bytes_read, we compute that
	binary_format="<HI"


	def __init__(self):
		self.ID=0
		self.length=0
		self.bytes_read=0

	def dump(self):
		print "ID: ", self.ID
		print "ID in hex: ", hex(self.ID)
		print "length: ", self.length
		print "bytes_read: ", self.bytes_read


def read_chunk(file, chunk):
		temp_data=file.read(struct.calcsize(chunk.binary_format))
		data=struct.unpack(chunk.binary_format, temp_data)
		chunk.ID=data[0]
		chunk.length=data[1]
		#update the bytes read function
		chunk.bytes_read=6

		#if debugging
		#chunk.dump()

def read_string(file):
	s=""
	index=0
	#print "reading a string"
	#read in the characters till we get a null character
	temp_data=file.read(struct.calcsize("c"))
	data=struct.unpack("c", temp_data)
	s=s+(data[0])
	#print "string: ",s
	while(ord(s[index])!=0):
		index+=1
		temp_data=file.read(struct.calcsize("c"))
		data=struct.unpack("c", temp_data)
		s=s+(data[0])
		#print "string: ",s
	return str(s)

######################################################
# EXPORT
######################################################
def save_3ds(filename):

	vert_list={}
	vert_count=0
	#open the current scene
	scene = Blender.Scene.GetCurrent()
	#if the user was lazy and didn't add a filename,
	#just name the files after the scene name
	#hope they weren't lazy there too.
	if filename=="model":
		filename=scene.getName()
		#basefilename and /path/filename are the same
		bfilename=filename
	else:
		#extract the actual filename from the
		#entire path of the /path/filename
		bfilename = os.path.basename(filename)

	#fill the chunks full of data
	primary=primary_chunk()
	#get all the objects in this scene
	object_list=Blender.Object.Get()
	#fill up the data structures with objects
	for obj in object_list:
		#create a new object chunk
		if obj.getType()=="Material":
			#primary.object_info.mat_chunks.append(material_chunk())
			pass
		elif obj.getType()=="Mesh":
			primary.obj_info.obj_chunks.append(object_chunk())
			#get the mesh data
			blender_mesh=obj.getData()
			#set the object name
			primary.obj_info.obj_chunks[len(primary.obj_info.obj_chunks)-1].name=obj.getName()

			#make a new mesh chunk object
			mesh=mesh_chunk()

			#fill verts in
			vert_count=0
			for vert in blender_mesh.verts:
				v = (vert.co)
				vert_key=(v[0], v[1], v[2])
				mesh.v_chunk.verts.append(vert_key)
				if not vert_list.has_key(vert_key):
					#print "didn't have the key"
					vert_list[vert_key] = vert_count  #save this info for the face lookup table
					vert_count = vert_count + 1

			print "vert_count: ", vert_count
			print "mesh.verts: ", len(blender_mesh.verts)


			#fill in faces
			for face in blender_mesh.faces:
				#is this a tri or a quad
				num_fv=len(face.v)
				
				#it's a tri
				if num_fv==3:
					indexes=[0]*3
					for i in range (0,3):
						#build the keys
						vert=(face.v[i].co)
						vert_key=(vert[0], vert[1], vert[2])
						#look up the key to get the index
						vert_index=vert_list[vert_key]
						indexes[i]=vert_index
					mesh.f_chunk.faces.append(indexes)  
					
				#it's a quad
				elif num_fv==4:
					indexes=[0]*4
					for i in range (0,4):
						#build the keys
						vert=(face.v[i].co)
						vert_key=(vert[0], vert[1], vert[2])
						#look up the key to get the index
						vert_index=vert_list[vert_key]
						indexes[i]=vert_index
						first_tri=(indexes[0], indexes[1], indexes[2])
						sec_tri=(indexes[2], indexes[3], indexes[0])
						
					mesh.f_chunk.faces.append(first_tri)  # 0,1,2
					mesh.f_chunk.faces.append(sec_tri)  # 2,3,0
				

			#fill in the UV info
			if blender_mesh.hasVertexUV():
				for vert in blender_mesh.verts:
					uv_coord=(vert.uvco)
					mesh.uv_chunk.uv.append((uv_coord[0], uv_coord[1]))

			elif blender_mesh.hasFaceUV():
				for face in blender_mesh.faces:
					num_fv=len(face.v)
					if num_fv==3:
						for i in range (0,3):
							uv_coord=(face.uv[i])
							mesh.uv_chunk.uv.append((uv_coord[0], uv_coord[1]))
					elif num_fv==4:
						for i in range (0,4):
							uv_coord=(face.uv[i])
							mesh.uv_chunk.uv.append((uv_coord[0], uv_coord[1]))


			#filled in our mesh, lets add it to the file
			primary.obj_info.obj_chunks[len(primary.obj_info.obj_chunks)-1].mesh_chunks.append(mesh)

	#check the size
	primary.get_size()
	#open the files up for writing
	file = open( filename, "wb" )
	#recursively write the stuff to file
	primary.write(file)
	file.close()

#***********************************************
# MAIN
#***********************************************

	
def my_callback(filename):
	if(filename.find('.3ds', -4) <= 0):
		filename += '.3ds'
	save_3ds(filename)

Blender.Window.FileSelector(my_callback, "Export 3DS")